@@ -8,12 +8,11 @@ from pathlib import Path
 from typing import Callable, List, Optional, Tuple, Union
 
 import questionary.prompts.common
-import yaml
 from prompt_toolkit.shortcuts import clear as clear_screen
 from prompt_toolkit.styles import Style
 from questionary import prompt
 
-from sparv import SourceStructureParser, Wildcard
+from sparv.api import SourceStructureParser, Wildcard
 from sparv.core import registry, paths, config, snake_utils
 from sparv.core.console import console
 
@@ -45,14 +44,20 @@ class Wizard:
         self.corpus_config = {}  # Dict holding the resulting config structure
         self.here = os.getcwd()
         self.config_file = os.path.join(self.here, paths.config_file)
+        self.wizard_from_module = defaultdict(set)
+        self.has_wildcard_choices = False
+        self.has_class_choices = False
 
         # Annotator info
+        self._annotator_max_len = 0
         self.snake_storage = None
         self.output_to_annotators = {}
         self.annotation_to_annotator = {}
         self.annotation_description = {}
         self.wildcard_annotations = []
         self.importers = defaultdict(dict)
+        self.exporters = defaultdict(dict)
+        self.exporters_with_wizard = []
 
         self.source_structure: Optional[SourceStructureParser] = None
 
@@ -63,9 +68,6 @@ class Wizard:
         self.annotation_description = {}
         registry.annotation_classes["config_classes"] = config.config.get("classes", {})
 
-        if not registry.modules:
-            registry.find_modules()
-
         self.snake_storage = snake_utils.SnakeStorage()
 
         for module_name in registry.modules:
@@ -88,6 +90,8 @@ class Wizard:
                             self.wildcard_annotations.append(annotation[0].original_name)
                 elif rule_storage.importer and rule_made:
                     self.importers[module_name][f_name] = rule_storage
+                elif rule_storage.exporter and rule_made:
+                    self.exporters[module_name][f_name] = rule_storage
 
     def q(self, questions: Union[List[dict], dict], clear: bool = False, save_prompt: bool = True):
         """Ask questions and handle interruptions.
@@ -174,10 +178,9 @@ class Wizard:
 
         return questions
 
-    def get_module_wizard(self, module_wizard: Tuple[Callable, list, bool],
-                          answers: Optional[dict] = None) -> List[dict]:
+    def get_module_wizard(self, module_wizard: Tuple[Callable, list, bool]) -> List[dict]:
         """Get wizard from module."""
-        args = [answers or {}]
+        args = [self.corpus_config]
         if module_wizard[2]:
             args.append(self.source_structure)
         questions = module_wizard[0](*args)
@@ -185,8 +188,8 @@ class Wizard:
 
     def save_config(self):
         """Save config to YAML file."""
-        with open("config.yaml", mode="w") as out_file:
-            yaml.dump({k: v for k, v in self.corpus_config.items() if not k.startswith("_")}, out_file)
+        with open("config.yaml", mode="w", encoding="utf-8") as out_file:
+            out_file.write(config.dump_config({k: v for k, v in self.corpus_config.items() if not k.startswith("_")}))
         print("Your corpus configuration has been saved as 'config.yaml'.")
 
     def load_config(self):
@@ -201,14 +204,18 @@ class Wizard:
                            "again. Select Y to continue or N to abort.".format(paths.config_file)
             }, clear=True, save_prompt=False)["answer"]
             if use_config_file:
-                with open(self.config_file) as f:
-                    self.corpus_config = yaml.load(f, Loader=yaml.FullLoader)
+                self.corpus_config = config.read_yaml(self.config_file)
                 config.load_config(self.config_file)
+                if not self.corpus_config:
+                    # If config is completely empty, treat as if config is missing
+                    return False
+                return True
             else:
                 sys.exit()
         else:
             # Load default config
             config.load_config(None)
+            return False
 
     def update_config(self, answers: dict):
         """Add answers to corpus_config dict and update full config."""
@@ -242,83 +249,67 @@ class Wizard:
     def run(self):
         """Run the Sparv corpus set-up wizard."""
         # Load default config and any existing corpus config
-        self.load_config()
+        existing_config = self.load_config()
 
         # Temporarily unset corpus language to allow all modules to be loaded
         language = config.get("metadata.language")
         config.set_value("metadata.language", None)
 
-        # Load all available annotators
-        self.update_annotators()
+        # Load all modules
+        registry.find_modules()
 
         # Restore language
         config.set_value("metadata.language", language)
 
+        # Update class registry with language specific classes
+        self.update_class_registry()
+
+        # Build annotation dictionaries
+        self.update_annotators()
+
         # Build module wizard index
         wizard_from_config = {}
-        wizard_from_module = defaultdict(set)
         for w in registry.wizards:
             for config_variable in w[1]:
                 wizard_from_config[config_variable] = w
-                wizard_from_module[config_variable.split(".")[0]].add(w)
+                self.wizard_from_module[config_variable.split(".")[0]].add(w)
 
-        # Initial question to check prerequisites
-        self.prerequisites()
-
-        # Start with metadata questions
-        questions = []
-        for w in wizard_from_module["metadata"]:
-            questions.extend(self.get_module_wizard(w))
-
-        self.update_config(self.q(questions, clear=True))
-
-        # Importer choice
-        questions = []
-        for w in wizard_from_module["import"]:
-            questions.extend(self.get_module_wizard(w))
-
-        self.update_config(self.q(questions))
-
-        # Now that the user has selected a language, update the class dict in registry...
-        for cls, targets in registry.all_module_classes[config.get("metadata.language")].items():
-            registry.annotation_classes["module_classes"][cls].extend(targets)
-
-        # ...and rebuild annotator list
-        self.update_annotators()
-
-        # Ask user if they want to scan source files
-        self.scan_source()
-
-        # Choose document annotation
-        self.select_document_annotation()
-
-        # Select source annotations to keep
-        questions = []
-        for w in wizard_from_module["export"]:
-            questions.extend(self.get_module_wizard(w))
-
-        self.update_config(self.q(questions))
-
-        # Parse annotations from existing config
+        # Annotations selected by the user
         selected_annotations = defaultdict(dict)
-        self.parse_config_annotations(selected_annotations)
 
-        # Select annotations
-        annotator_max_len = self.select_annotations(selected_annotations)
+        if not existing_config:
+            # Initial question to check prerequisites
+            self.prerequisites()
 
-        # Select classes if needed
-        has_class_choices = self.select_classes(selected_annotations)
+            # Start with metadata questions
+            self.metadata_questions()
 
-        # Select wildcards if needed
-        has_wildcard_choices = self.select_wildcards(selected_annotations)
+            # Questions related to the source files
+            self.source_questions()
+
+            # Select annotations
+            self.select_annotations(selected_annotations)
+
+            # Select default export formats
+            self.select_exports()
+        else:
+            # Parse annotations from existing config
+            self.parse_config_annotations(selected_annotations)
+            self.select_classes(selected_annotations)
+            self.select_wildcards(selected_annotations)
 
         # Add selected annotations to config
         self.update_export_annotations(selected_annotations)
 
+        # Check for configurable export modules
+        export_modules = sorted(self.exporters.keys())
+        self.exporters_with_wizard = [m for m in export_modules if self.wizard_from_module.get(m)]
+
         # Set config variables
-        self.edit_config(selected_annotations, annotator_max_len)
+        self.edit_config(selected_annotations)
 
         # We're done collecting the required data. Let the user edit further if they want to.
+        preselected = None
         while True:
             choices = [
                 {
@@ -326,63 +317,156 @@ class Wizard:
                     "value": "done"
                 }
             ]
-
-            if has_class_choices:
+            if existing_config:
+                choices.extend([
+                    {
+                        "name": "Edit metadata configurations",
+                        "value": "metadata"
+                    },
+                    {
+                        "name": "Edit settings related to the source files",
+                        "value": "source"
+                    },
+                    {
+                        "name": "Select annotations",
+                        "value": "annotations"
+                    },
+                    {
+                        "name": "Select default export formats",
+                        "value": "exports"
+                    }
+                ])
+            if self.has_class_choices:
                 choices.append({
                     "name": "Edit class choices",
                     "value": "class"
                 })
-            if has_wildcard_choices:
+            if self.has_wildcard_choices:
                 choices.append({
                     "name": "Edit wildcard references",
                     "value": "wildcard"
                 })
             choices.append({
                 "name": "Edit annotator configurations",
-                "value": "config"
+                "value": "annotator_config"
             })
+            if self.exporters_with_wizard:
+                choices.append({
+                    "name": "Edit exporter configurations",
+                    "value": "exporter_configs"
+                })
 
-            choice = self.q({
+            message_new_config = ("All the necessary data has been collected, but you may do further customization by "
+                                  "selecting one of the options below.")
+            message_prev_config = "What would you like to do?"
+
+            choice = self.q(dict({
                 "name": "choice",
                 "type": "select",
                 "choices": choices,
-                "message": "All the necessary data has been collected, but you may do further customization by "
-                           "selecting one of the options below."
-            }, clear=True)["choice"]
+                "message": message_new_config if not existing_config else message_prev_config
+            }, **{"default": preselected} if preselected else {}), clear=True)["choice"]
 
             if choice == "done":
                 break
+            elif choice == "exporter_configs":
+                self.configure_exporters()
             elif choice == "class":
-                has_class_choices = self.select_classes(selected_annotations, always_ask=True)
+                self.select_classes(selected_annotations, always_ask=True)
             elif choice == "wildcard":
-                has_wildcard_choices = self.select_wildcards(selected_annotations, always_ask=True)
+                self.select_wildcards(selected_annotations, always_ask=True)
                 self.update_export_annotations(selected_annotations)
-            elif choice == "config":
-                self.edit_config(selected_annotations, annotator_max_len, show_optional=True)
+            elif choice == "annotator_config":
+                self.edit_config(selected_annotations, show_optional=True)
+            elif choice == "metadata":
+                self.metadata_questions()
+            elif choice == "source":
+                self.source_questions()
+            elif choice == "annotations":
+                self.select_annotations(selected_annotations)
+                self.update_export_annotations(selected_annotations)
+            elif choice == "exports":
+                self.select_exports()
+
+            preselected = choices[[c["value"] for c in choices].index(choice)]
 
         self.save_config()
 
-    def select_document_annotation(self):
-        """Ask user for document annotation."""
-        doc_annotation = self.q(self.set_defaults({
+    @property
+    def annotator_max_len(self):
+        """Calculate max length of annotator names."""
+        if not self._annotator_max_len:
+            self._annotator_max_len = max(
+                len(module + f_name) + 1 for module in self.snake_storage.all_annotators for f_name in
+                self.snake_storage.all_annotators[module])
+        return self._annotator_max_len
+
+    @staticmethod
+    def update_class_registry():
+        """Update class registry with language specific classes."""
+        for cls, targets in registry.all_module_classes[config.get("metadata.language")].items():
+            for target in targets:
+                if target not in registry.annotation_classes["module_classes"][cls]:
+                    registry.annotation_classes["module_classes"][cls].append(target)
+
+    def metadata_questions(self):
+        """Run metadata wizard."""
+        questions = []
+        for w in self.wizard_from_module["metadata"]:
+            questions.extend(self.get_module_wizard(w))
+        self.update_config(self.q(questions, clear=True))
+        # Split language into language and variety if necessary and save in config
+        langcode, _, suffix = config.get("metadata.language", "").partition("-")
+        if suffix:
+            config.set_value("metadata.language", langcode, config_dict=self.corpus_config)
+            config.set_value("metadata.variety", suffix, config_dict=self.corpus_config)
+
+        # Now that we know the language, update the class dict in registry...
+        self.update_class_registry()
+        # ...and rebuild annotator list
+        self.update_annotators()
+
+    def source_questions(self):
+        """As questions related to the source files."""
+        # Importer choice
+        questions = []
+        for w in self.wizard_from_module["import"]:
+            questions.extend(self.get_module_wizard(w))
+        self.update_config(self.q(questions, clear=True))
+
+        # Ask user if they want to scan source files
+        self.scan_source()
+
+        # Choose text annotation
+        self.select_text_annotation()
+
+        # Select source annotations to keep
+        questions = []
+        for w in self.wizard_from_module["export"]:
+            questions.extend(self.get_module_wizard(w))
+        self.update_config(self.q(questions))
+
+    def select_text_annotation(self):
+        """Ask user for text annotation."""
+        text_annotation = self.q(self.set_defaults({
             "type": "select",
-            "name": "import.document_annotation",
+            "name": "import.text_annotation",
             "message": "What is the name of the existing annotation in your source files that encapsulates a "
-                       "'document'? This is the text unit for which all document level annotations will apply. "
-                       "Usually, no text should exist outside of this annotation.",
+                       "'text'? This is the text unit for which all text level annotations will apply. "
+                       "Usually, no text content should exist outside of this annotation.",
             "choices": self.source_structure.get_plain_annotations(self.corpus_config) + [{
                 "name": "Enter manually",
                 "value": "__sparv_manual_entry"
             }]
         }))
-        if doc_annotation["import.document_annotation"] == "__sparv_manual_entry":
-            doc_annotation = self.q({
+        if text_annotation["import.text_annotation"] == "__sparv_manual_entry":
+            text_annotation = self.q({
                 "type": "text",
-                "name": "import.document_annotation",
+                "name": "import.text_annotation",
                 "message": "Annotation name, e.g. 'text' or  ̈́document':",
                 "validate": lambda x: bool(re.match(r"^\S+$", x))
             })
-        self.update_config(doc_annotation)
+        self.update_config(text_annotation)
 
     def scan_source(self):
         """Create a SourceStructureParser instance and offer to scan source files if possible."""
@@ -427,7 +511,7 @@ class Wizard:
                     annotations.append(value)
         config.set_value("export.annotations", annotations, config_dict=self.corpus_config)
 
-    def edit_config(self, selected_annotations, annotator_max_len, show_optional: bool = False):
+    def edit_config(self, selected_annotations, show_optional: bool = False):
         """Ask the user for required config variables."""
 
         def get_dependencies(module, f):
@@ -492,7 +576,7 @@ class Wizard:
                             selected_annotations[
                                 module].get(a) else "   ",
                             self.snake_storage.all_annotators[module][a]["rule"].description,
-                            width=annotator_max_len + (
+                            width=self.annotator_max_len + (
                                 0 if not self.snake_storage.all_annotators[module][a][
                                     "rule"].configs else 2)),
                         "value": (module, a),
@@ -559,12 +643,9 @@ class Wizard:
                             config.set_value(config_choice, config_value)
                             config.set_value(config_choice, config_value, config_dict=self.corpus_config)
 
-    def select_wildcards(self, selected_annotations, always_ask: bool = False) -> bool:
-        """Find any annotations with wildcards and prompt the user to define values for these.
-
-        Returns True if there are wildcards, otherwise False.
-        """
-        has_wildcards = False
+    def select_wildcards(self, selected_annotations, always_ask: bool = False):
+        """Find any annotations with wildcards and prompt the user to define values for these."""
+        self.has_wildcard_choices = False
         full_annotations = set()
         plain_annotations = set()
         attributes = set()
@@ -589,7 +670,7 @@ class Wizard:
                 if selected_annotations[module][f_name]:
                     wildcards = self.snake_storage.all_annotators[module][f_name]["rule"].wildcards
                     if wildcards:
-                        has_wildcards = True
+                        self.has_wildcard_choices = True
                         wc_dict = {wc.name: wc for wc in wildcards}
                         wildcard_max_len = max([len(wc.name) for wc in wildcards])
                         selected_wildcards = {}
@@ -675,21 +756,17 @@ class Wizard:
                         for annotation in selected_annotations[module][f_name]:
                             if "{" in annotation["annotation"]:
                                 annotation["wildcards"] = selected_wildcards
-        return has_wildcards
 
-    def select_classes(self, selected_annotations, always_ask: bool = False) -> bool:
-        """Find if any dependencies are using classes with multiple options. If so, ask the user which to use.
-
-        Returns True if there are multiple options, otherwise False.
-        """
-        has_choices = False
+    def select_classes(self, selected_annotations, always_ask: bool = False):
+        """Find if any dependencies are using classes with multiple options. If so, ask the user which to use."""
+        self.has_class_choices = False
         available_classes = registry.annotation_classes["module_classes"]
         for module in selected_annotations:
             for f_name in selected_annotations[module]:
                 if selected_annotations[module][f_name]:
                     for cls in self.snake_storage.all_annotators[module][f_name]["rule"].classes:
                         if len(available_classes.get(cls, [])) > 1:
-                            has_choices = True
+                            self.has_class_choices = True
                             if not always_ask and config.get(f"classes.{cls}"):
                                 continue
                             max_cls_len = max(map(len, available_classes[cls]))
@@ -713,19 +790,18 @@ class Wizard:
 
                             self.update_config(selected_class)
                             self.update_annotators()
-        return has_choices
 
     def select_annotations(self, selected_annotations) -> int:
         """Let the user select annotators and annotations to use."""
-        max_len = max(
-            len(module + f_name) + 1 for module in self.snake_storage.all_annotators for f_name in
-            self.snake_storage.all_annotators[module])
         annotator_choice = None
         while True:
             annotators = []
             preselected = None
             for module in self.snake_storage.all_annotators:
                 for a in self.snake_storage.all_annotators[module]:
+                    # Skip utility annotators
+                    if self.snake_storage.all_custom_annotators.get(module, {}).get(a):
+                        continue
                     annotators.append({
                         "name": "{:{width}} {}  {}".format(
                             "{}:{}".format(module, a),
@@ -733,7 +809,7 @@ class Wizard:
                             selected_annotations[
                                 module].get(a) else "   ",
                             self.snake_storage.all_annotators[module][a]["rule"].description,
-                            width=max_len),
+                            width=self.annotator_max_len),
                         "value": (module, a),
                         "short": "{}:{}".format(module, a)
                     })
@@ -781,7 +857,43 @@ class Wizard:
                     selected_annotations[module][annotator].append({"annotation": annotation})
             else:
                 break
-        return max_len
+        # Select classes if needed
+        self.select_classes(selected_annotations)
+        # Select wildcards if needed
+        self.select_wildcards(selected_annotations)
+
+    def select_exports(self):
+        """Ask user to choose default export formats."""
+        export_formats = [f"{k}:{e}" for k, v in self.exporters.items() for e in v]
+        default_exports = self.q(self.set_defaults({
+            "type": "checkbox",
+            "name": "export.default",
+            "message": "What export formats do you want to generate when running 'sparv run'?",
+            "choices": export_formats
+        }), clear=True)
+        self.update_config(default_exports)
+
+    def configure_exporters(self):
+        """Ask user to configure exporters."""
+        while True:
+            exporter_choice = self.q({
+                "type": "select",
+                "name": "exporter",
+                "message": "Which exporter would you like to configure?",
+                "choices": [
+                    {
+                        "name": DONE,
+                        "value": "_done"
+                    },
+                ] + self.exporters_with_wizard
+            }, clear=True)["exporter"]
+            if exporter_choice != "_done":
+                questions = []
+                for w in self.wizard_from_module.get(exporter_choice):
+                    questions.extend(self.get_module_wizard(w))
+                self.update_config(self.q(questions))
+            else:
+                break
 
     def parse_config_annotations(self, selected_annotations) -> None:
         """Parse selected annotations from existing config."""
@@ -810,7 +922,8 @@ class Wizard:
                     val["wildcards"] = wildcards
                 selected_annotations[module_name][f_name].append(val)
             else:
-                print(f"Could not parse the annotation '{annotation}'.")
+                print(f"Could not parse the annotation '{annotation}'. Either it's invalid, or it's not supported by "
+                      "the wizard.")
                 sys.exit(1)
 
 
