@@ -1,18 +1,19 @@
 """Tools for exporting, encoding and aligning corpora for Corpus Workbench."""
 
-import logging
 import os
 import re
+from collections import OrderedDict
 from glob import glob
 from pathlib import Path
 from typing import Optional
 
-import sparv.util as util
-from sparv import (AllDocuments, Annotation, AnnotationAllDocs, Config, Corpus, Document, Export, ExportAnnotations,
-                   ExportInput, SourceAnnotations, exporter)
-from sparv.core import paths
+from sparv.api import (AllSourceFilenames, Annotation, AnnotationAllSourceFiles, Config, Corpus, SourceFilename, Export,
+                       ExportAnnotations, ExportAnnotationNames, ExportInput, SourceAnnotations,
+                       SourceAnnotationsAllSourceFiles, SparvErrorMessage, exporter, get_logger, util)
 
-log = logging.getLogger(__name__)
+logger = get_logger(__name__)
+
+CWB_MAX_LINE_LEN = 65534
 
 
 @exporter("VRT export", config=[
@@ -21,8 +22,8 @@ log = logging.getLogger(__name__)
                        "included by default."),
     Config("cwb.annotations", description="Sparv annotations to include.")
 ])
-def vrt(doc: Document = Document(),
-        out: Export = Export("vrt/{doc}.vrt"),
+def vrt(source_file: SourceFilename = SourceFilename(),
+        out: Export = Export("cwb.vrt/{file}.vrt"),
         token: Annotation = Annotation("<token>"),
         word: Annotation = Annotation("[export.word]"),
         annotations: ExportAnnotations = ExportAnnotations("cwb.annotations"),
@@ -33,7 +34,7 @@ def vrt(doc: Document = Document(),
     """Export annotations to vrt.
 
     - annotations: list of elements:attributes (annotations) to include.
-    - source_annotations: list of elements:attributes from the original document
+    - source_annotations: list of elements:attributes from the original source file
       to be kept. If not specified, everything will be kept.
     """
     # Create export dir
@@ -43,26 +44,28 @@ def vrt(doc: Document = Document(),
     word_annotation = list(word.read())
 
     # Get annotation spans, annotations list etc.
-    annotation_list, token_attributes, export_names = util.get_annotation_names(annotations, source_annotations,
-                                                                                doc=doc, token_name=token.name,
-                                                                                remove_namespaces=remove_namespaces,
-                                                                                sparv_namespace=sparv_namespace,
-                                                                                source_namespace=source_namespace)
-    span_positions, annotation_dict = util.gather_annotations(annotation_list, export_names, doc=doc)
+    annotation_list, token_attributes, export_names = util.export.get_annotation_names(
+        annotations, source_annotations, source_file=source_file, token_name=token.name,
+        remove_namespaces=remove_namespaces, sparv_namespace=sparv_namespace, source_namespace=source_namespace)
+    if token not in annotation_list:
+        logger.warning("The 'cwb:vrt' export requires the <token> annotation for the output to include "
+                       "the source text. Make sure to add <token> to the list of export annotations.")
+    span_positions, annotation_dict = util.export.gather_annotations(annotation_list, export_names,
+                                                                     source_file=source_file)
     vrt_data = create_vrt(span_positions, token.name, word_annotation, token_attributes, annotation_dict,
                           export_names)
 
     # Write result to file
-    with open(out, "w") as f:
+    with open(out, "w", encoding="utf-8") as f:
         f.write(vrt_data)
-    log.info("Exported: %s", out)
+    logger.info("Exported: %s", out)
 
 
 @exporter("Scrambled VRT export", config=[
     Config("cwb.scramble_on", description="Annotation to use for scrambling.")
 ])
-def vrt_scrambled(doc: Document = Document(),
-                  out: Export = Export("vrt_scrambled/{doc}.vrt"),
+def vrt_scrambled(source_file: SourceFilename = SourceFilename(),
+                  out: Export = Export("cwb.vrt_scrambled/{file}.vrt"),
                   chunk: Annotation = Annotation("[cwb.scramble_on]"),
                   chunk_order: Annotation = Annotation("[cwb.scramble_on]:misc.number_random"),
                   token: Annotation = Annotation("<token>"),
@@ -73,117 +76,117 @@ def vrt_scrambled(doc: Document = Document(),
                   sparv_namespace: str = Config("export.sparv_namespace"),
                   source_namespace: str = Config("export.source_namespace")):
     """Export annotations to vrt in scrambled order."""
+    logger.progress(total=6)
     # Get annotation spans, annotations list etc.
-    annotation_list, token_attributes, export_names = util.get_annotation_names(annotations, source_annotations,
-                                                                                doc=doc, token_name=token.name,
-                                                                                remove_namespaces=remove_namespaces,
-                                                                                sparv_namespace=sparv_namespace,
-                                                                                source_namespace=source_namespace)
+    annotation_list, token_attributes, export_names = util.export.get_annotation_names(
+        annotations, source_annotations, source_file=source_file, token_name=token.name,
+        remove_namespaces=remove_namespaces, sparv_namespace=sparv_namespace, source_namespace=source_namespace)
+    logger.progress()
+    if token not in annotation_list:
+        logger.warning("The 'cwb:vrt_scrambled' export requires the <token> annotation for the output to include "
+                       "the source text. Make sure to add <token> to the list of export annotations.")
     if chunk not in annotation_list:
-        raise util.SparvErrorMessage(
+        raise SparvErrorMessage(
             "The annotation used for scrambling ({}) needs to be included in the output.".format(chunk))
-    span_positions, annotation_dict = util.gather_annotations(annotation_list, export_names, doc=doc,
-                                                              split_overlaps=True)
+    span_positions, annotation_dict = util.export.gather_annotations(annotation_list, export_names,
+                                                                     source_file=source_file, split_overlaps=True)
+    logger.progress()
 
-    # Read words and document ID
+    # Read words and scramble order
     word_annotation = list(word.read())
     chunk_order_data = list(chunk_order.read())
 
-    # Reorder chunks and open/close tags in correct order
-    new_span_positions = util.scramble_spans(span_positions, chunk.name, chunk_order_data)
+    logger.progress()
 
+    # Reorder chunks and open/close tags in correct order
+    new_span_positions = util.export.scramble_spans(span_positions, chunk.name, chunk_order_data)
+    logger.progress()
     # Make vrt format
     vrt_data = create_vrt(new_span_positions, token.name, word_annotation, token_attributes, annotation_dict,
                           export_names)
-
+    logger.progress()
     # Create export dir
     os.makedirs(os.path.dirname(out), exist_ok=True)
 
     # Write result to file
-    with open(out, "w") as f:
+    with open(out, "w", encoding="utf-8") as f:
         f.write(vrt_data)
-    log.info("Exported: %s", out)
+    logger.info("Exported: %s", out)
+    logger.progress()
 
 
 @exporter("CWB encode", order=2, config=[
-    Config("cwb.corpus_registry", default=paths.corpus_registry, description="Path to CWB registry directory"),
-    Config("cwb.cwb_datadir", default=paths.cwb_datadir, description="Path to CWB data directory"),
     Config("cwb.bin_path", default="", description="Path to directory containing the CWB executables"),
-    Config("cwb.encoding", default=paths.cwb_encoding, description="Encoding to use"),
+    Config("cwb.encoding", default="utf8", description="Encoding to use"),
     Config("cwb.skip_compression", False, description="Whether to skip compression"),
     Config("cwb.skip_validation", False, description="Whether to skip validation")
 ])
 def encode(corpus: Corpus = Corpus(),
-           annotations: ExportAnnotations = ExportAnnotations("cwb.annotations", is_input=False),
-           source_annotations: SourceAnnotations = SourceAnnotations("cwb.source_annotations"),
-           docs: AllDocuments = AllDocuments(),
-           words: AnnotationAllDocs = AnnotationAllDocs("[export.word]"),
-           vrtfiles: ExportInput = ExportInput("vrt/{doc}.vrt", all_docs=True),
-           out: Export = Export("[cwb.corpus_registry]/[metadata.id]", absolute_path=True),
-           out_marker: Export = Export("[cwb.cwb_datadir]/[metadata.id]/.original_marker", absolute_path=True),
-           token: AnnotationAllDocs = AnnotationAllDocs("<token>"),
+           annotations: ExportAnnotationNames = ExportAnnotationNames("cwb.annotations"),
+           source_annotations: SourceAnnotationsAllSourceFiles = SourceAnnotationsAllSourceFiles(
+               "cwb.source_annotations"),
+           source_files: AllSourceFilenames = AllSourceFilenames(),
+           words: AnnotationAllSourceFiles = AnnotationAllSourceFiles("[export.word]"),
+           vrtfiles: ExportInput = ExportInput("cwb.vrt/{file}.vrt", all_files=True),
+           out_registry: Export = Export("cwb.encoded/registry/[metadata.id]"),
+           out_marker: Export = Export("cwb.encoded/data/.marker"),
+           token: AnnotationAllSourceFiles = AnnotationAllSourceFiles("<token>"),
            bin_path: Config = Config("cwb.bin_path"),
            encoding: str = Config("cwb.encoding"),
-           datadir: str = Config("cwb.cwb_datadir"),
-           registry: str = Config("cwb.corpus_registry"),
            remove_namespaces: bool = Config("export.remove_module_namespaces", False),
            sparv_namespace: str = Config("export.sparv_namespace"),
            source_namespace: str = Config("export.source_namespace"),
            skip_compression: Optional[bool] = Config("cwb.skip_compression"),
            skip_validation: Optional[bool] = Config("cwb.skip_validation")):
-    """Do cwb encoding with vrt files in original order."""
-    cwb_encode(corpus, annotations, source_annotations, docs, words, vrtfiles, out, out_marker, token.name,
-               bin_path, encoding, datadir, registry, remove_namespaces, sparv_namespace, source_namespace,
+    """Encode CWB corpus from VRT files."""
+    cwb_encode(corpus, annotations, source_annotations, source_files, words, vrtfiles, out_registry, out_marker,
+               token.name, bin_path, encoding, remove_namespaces, sparv_namespace, source_namespace,
                skip_compression, skip_validation)
 
 
 @exporter("CWB encode, scrambled", order=1)
 def encode_scrambled(corpus: Corpus = Corpus(),
-                     annotations: ExportAnnotations = ExportAnnotations("cwb.annotations", is_input=False),
-                     source_annotations: SourceAnnotations = SourceAnnotations("cwb.source_annotations"),
-                     docs: AllDocuments = AllDocuments(),
-                     words: AnnotationAllDocs = AnnotationAllDocs("[export.word]"),
-                     vrtfiles: ExportInput = ExportInput("vrt_scrambled/{doc}.vrt", all_docs=True),
-                     out: Export = Export("[cwb.corpus_registry]/[metadata.id]", absolute_path=True),
-                     out_marker: Export = Export("[cwb.cwb_datadir]/[metadata.id]/.scrambled_marker",
-                                                 absolute_path=True),
-                     token: AnnotationAllDocs = AnnotationAllDocs("<token>"),
+                     annotations: ExportAnnotationNames = ExportAnnotationNames("cwb.annotations"),
+                     source_annotations: SourceAnnotationsAllSourceFiles = SourceAnnotationsAllSourceFiles(
+                         "cwb.source_annotations"),
+                     source_files: AllSourceFilenames = AllSourceFilenames(),
+                     words: AnnotationAllSourceFiles = AnnotationAllSourceFiles("[export.word]"),
+                     vrtfiles: ExportInput = ExportInput("cwb.vrt_scrambled/{file}.vrt", all_files=True),
+                     out_registry: Export = Export("cwb.encoded_scrambled/registry/[metadata.id]"),
+                     out_marker: Export = Export("cwb.encoded_scrambled/data/.scrambled_marker"),
+                     token: AnnotationAllSourceFiles = AnnotationAllSourceFiles("<token>"),
                      bin_path: Config = Config("cwb.bin_path"),
                      encoding: str = Config("cwb.encoding"),
-                     datadir: str = Config("cwb.cwb_datadir"),
-                     registry: str = Config("cwb.corpus_registry"),
                      remove_namespaces: bool = Config("export.remove_module_namespaces", False),
                      sparv_namespace: str = Config("export.sparv_namespace"),
                      source_namespace: str = Config("export.source_namespace"),
                      skip_compression: Optional[bool] = Config("cwb.skip_compression"),
                      skip_validation: Optional[bool] = Config("cwb.skip_validation")):
-    """Do cwb encoding with vrt files in scrambled order."""
-    cwb_encode(corpus, annotations, source_annotations, docs, words, vrtfiles, out, out_marker, token.name,
-               bin_path, encoding, datadir, registry, remove_namespaces, sparv_namespace, source_namespace,
+    """Encode CWB corpus from scrambled VRT files."""
+    cwb_encode(corpus, annotations, source_annotations, source_files, words, vrtfiles, out_registry, out_marker,
+               token.name, bin_path, encoding, remove_namespaces, sparv_namespace, source_namespace,
                skip_compression, skip_validation)
 
 
-def cwb_encode(corpus, annotations, source_annotations, docs, words, vrtfiles, out, out_marker, token_name: str,
-               bin_path, encoding, datadir, registry, remove_namespaces, sparv_namespace, source_namespace,
+def cwb_encode(corpus, annotations, source_annotations, source_files, words, vrtfiles, out_registry, out_marker,
+               token_name: str, bin_path, encoding, remove_namespaces, sparv_namespace, source_namespace,
                skip_compression, skip_validation):
     """Encode a number of vrt files, by calling cwb-encode."""
-    assert datadir, "CWB_DATADIR not specified"
-    assert registry, "CORPUS_REGISTRY not specified"
+    if not corpus.strip():
+        raise SparvErrorMessage("metadata.id needs to be set.")
 
     # Get vrt files
-    vrtfiles = [vrtfiles.replace("{doc}", doc) for doc in docs]
+    vrtfiles = [vrtfiles.replace("{file}", file) for file in source_files]
     vrtfiles.sort()
 
     # Word annotation should always be included in CWB export
     annotations.insert(0, (words, None))
 
     # Get annotation names
-    annotation_list, token_attributes, export_names = util.get_annotation_names(annotations, source_annotations,
-                                                                                docs=docs, token_name=token_name,
-                                                                                remove_namespaces=remove_namespaces,
-                                                                                sparv_namespace=sparv_namespace,
-                                                                                source_namespace=source_namespace,
-                                                                                keep_struct_names=True)
+    annotation_list, token_attributes, export_names = util.export.get_annotation_names(
+        annotations, source_annotations, source_files=source_files, token_name=token_name,
+        remove_namespaces=remove_namespaces, sparv_namespace=sparv_namespace, source_namespace=source_namespace,
+        keep_struct_names=True)
 
     # Get VRT columns
     token_attributes = [(token_name + ":" + i) for i in token_attributes]
@@ -196,13 +199,17 @@ def cwb_encode(corpus, annotations, source_annotations, docs, words, vrtfiles, o
                           not a.annotation_name == token_name]
     structs = parse_structural_attributes(struct_annotations)
 
-    corpus_registry = os.path.join(registry, corpus)
-    corpus_datadir = os.path.join(datadir, corpus)
-    util.system.clear_directory(corpus_datadir)
+    data_dir = Path(out_marker).resolve().parent
+    registry_dir = Path(out_registry).resolve().parent
+    registry_file = Path(out_registry).resolve()
+
+    # Create export dirs
+    data_dir.mkdir(exist_ok=True)
+    registry_dir.mkdir(exist_ok=True)
 
     encode_args = ["-s", "-p", "-",
-                   "-d", corpus_datadir,
-                   "-R", corpus_registry,
+                   "-d", data_dir,
+                   "-R", registry_file,
                    "-c", encoding,
                    "-x"
                    ]
@@ -214,38 +221,43 @@ def cwb_encode(corpus, annotations, source_annotations, docs, words, vrtfiles, o
         if col != "-":
             encode_args += ["-P", col]
     for struct, attrs in structs:
-        attrs2 = "+".join(attr for attr, _n in attrs if not attr == util.UNDEF)
+        attrs2 = "+".join(attrs)
         if attrs2:
             attrs2 = "+" + attrs2
+        # ":0" is added to the s-attribute name to enable nesting support in cwb-encode
         encode_args += ["-S", "%s:0%s" % (struct, attrs2)]
 
-    util.system.call_binary(os.path.join(bin_path, "cwb-encode"), encode_args, verbose=True)
+    _, stderr = util.system.call_binary(os.path.join(bin_path, "cwb-encode"), encode_args)
+    if stderr:
+        logger.warning(stderr.decode().strip())
     # Use xargs to avoid "Argument list too long" problems
-    # util.system.call_binary(os.path.join(bin_path, "cwb-encode"), raw_command="cat %s | xargs cat | %%s %s" % (vrtfiles, " ".join(encode_args)), use_shell=True)
+    # util.system.call_binary(os.path.join(bin_path, "cwb-encode"),
+    #                         raw_command="cat %s | xargs cat | %%s %s" % (vrtfiles, " ".join(encode_args)),
+    #                         use_shell=True)
 
-    index_args = ["-V", "-r", registry, corpus.upper()]
+    index_args = ["-V", "-r", registry_dir, corpus.upper()]
     util.system.call_binary(os.path.join(bin_path, "cwb-makeall"), index_args)
-    log.info("Encoded and indexed %d columns, %d structs", len(columns), len(structs))
+    logger.info("Encoded and indexed %d columns, %d structs", len(columns), len(structs))
 
     if not skip_compression:
-        log.info("Compressing corpus files...")
-        compress_args = ["-A", corpus.upper()]
+        logger.info("Compressing corpus files...")
+        compress_args = ["-A", "-r", registry_dir, corpus.upper()]
         if skip_validation:
             compress_args.insert(0, "-T")
-            log.info("Skipping validation")
+            logger.info("Skipping validation")
         # Compress token stream
         util.system.call_binary(os.path.join(bin_path, "cwb-huffcode"), compress_args)
-        log.info("Removing uncompressed token stream...")
-        for f in glob(os.path.join(corpus_datadir, "*.corpus")):
+        logger.info("Removing uncompressed token stream...")
+        for f in glob(os.path.join(data_dir, "*.corpus")):
             os.remove(f)
         # Compress index files
         util.system.call_binary(os.path.join(bin_path, "cwb-compress-rdx"), compress_args)
-        log.info("Removing uncompressed index files...")
-        for f in glob(os.path.join(corpus_datadir, "*.corpus.rev")):
+        logger.info("Removing uncompressed index files...")
+        for f in glob(os.path.join(data_dir, "*.corpus.rev")):
             os.remove(f)
-        for f in glob(os.path.join(corpus_datadir, "*.corpus.rdx")):
+        for f in glob(os.path.join(data_dir, "*.corpus.rdx")):
             os.remove(f)
-        log.info("Compression done.")
+        logger.info("Compression done.")
 
     # Write marker file
     Path(out_marker).touch()
@@ -257,7 +269,7 @@ def cwb_align(corpus, other, link, aligndir="annotations/align", bin_path="",
     """Align 'corpus' with 'other' corpus, using the 'link' annotation for alignment."""
     os.makedirs(aligndir, exist_ok=True)
     alignfile = os.path.join(aligndir, corpus + ".align")
-    log.info("Aligning %s <-> %s", corpus, other)
+    logger.info("Aligning %s <-> %s", corpus, other)
 
     try:
         [(link_name, [(link_attr, _path)])] = parse_structural_attributes(link)
@@ -268,34 +280,34 @@ def cwb_align(corpus, other, link, aligndir="annotations/align", bin_path="",
     # Align linked chunks
     args = ["-v", "-o", alignfile, "-V", link_attr, corpus, other, link_name]
     result, _ = util.system.call_binary(os.path.join(bin_path, "cwb-align"), args, encoding=encoding)
-    with open(alignfile + ".result", "w") as F:
+    with open(alignfile + ".result", "w", encoding="utf-8") as F:
         print(result, file=F)
     _, lastline = result.rsplit("Alignment complete.", 1)
-    log.info("%s", lastline.strip())
+    logger.info("%s", lastline.strip())
     if " 0 alignment" in lastline.strip():
-        log.warning("No alignment regions created")
-    log.info("Alignment file/result: %s/.result", alignfile)
+        logger.warning("No alignment regions created")
+    logger.info("Alignment file/result: %s/.result", alignfile)
 
     # Add alignment parameter to registry
     # cwb-regedit is not installed by default, so we skip it and modify the regfile directly instead:
     regfile = os.path.join(os.environ["CORPUS_REGISTRY"], corpus)
-    with open(regfile) as F:
+    with open(regfile, encoding="utf-8") as F:
         skip_align = ("ALIGNED %s" % other) in F.read()
 
     if not skip_align:
-        with open(regfile, "a") as F:
+        with open(regfile, "a", encoding="utf-8") as F:
             print(file=F)
             print("# Added by cwb.py", file=F)
             print("ALIGNED", other, file=F)
-        log.info("Added alignment to registry: %s", regfile)
+        logger.info("Added alignment to registry: %s", regfile)
     # args = [corpus, ":add", ":a", other]
     # result, _ = util.system.call_binary(os.path.join(bin_path, "cwb-regedit"), args)
-    # log.info("%s", result.strip())
+    # logger.info("%s", result.strip())
 
     # Encode the alignments into CWB
     args = ["-v", "-D", alignfile]
     result, _ = util.system.call_binary(os.path.join(bin_path, "cwb-align-encode"), args, encoding=encoding)
-    log.info("%s", result.strip())
+    logger.info("%s", result.strip())
 
 
 ################################################################################
@@ -349,46 +361,42 @@ def make_token_line(word, token, token_attributes, annotation_dict, index):
     line = [word.replace(" ", "_").replace("&", "&amp;").replace("<", "&lt;").replace(">", "&gt;")]
     for attr in token_attributes:
         if attr not in annotation_dict[token]:
-            attr_str = util.UNDEF
+            attr_str = util.constants.UNDEF
         else:
             attr_str = annotation_dict[token][attr][index]
         line.append(
             attr_str.replace(" ", "_").replace("/", "").replace("&", "&amp;").replace("<", "&lt;").replace(">", "&gt;"))
-    line = "\t".join(line)
-    return util.remove_control_characters(line)
+    line = util.misc.remove_control_characters("\t".join(line))
+
+    # Send warning if line exceeds the max line length in CWB
+    if len(line) > CWB_MAX_LINE_LEN:
+        if "misc.head" in token_attributes or "misc.tail" in token_attributes:
+            logger.warning(f"Found a line that exceeds the max line length in CWB ({CWB_MAX_LINE_LEN}). "
+                "If this leads to a crash in CWB encode you could try setting the 'misc.head_tail_max_length' config "
+                "variable to a lower value. In that case you will need to re-run the misc.head and misc.tail analyses.")
+        else:
+            logger.warning(f"Found a line that exceeds the max line length in CWB ({CWB_MAX_LINE_LEN}). "
+                           "This may lead to a crash in CWB encode.")
+
+    return line
 
 
 def parse_structural_attributes(structural_atts):
-    """Parse a list of annotations (element:attribute) into a list of tuples.
-
-    >>> parse_structural_attributes("s - text:title text:author")
-    [('s', [('__UNDEF__', 0)]), ('text', [('title', 2), ('author', 3)])]
-    """
-    if isinstance(structural_atts, str):
-        structural_atts = structural_atts.split()
-    structs = {}
-    order = []
-    for n, struct in enumerate(structural_atts):
-
-        # From the CWB documentation: "By convention, all attribute names must be lowercase
-        # (more precisely, they may only contain the characters a-z, 0-9, -, and _, and may not start with a digit)"
-        assert not struct or struct == "-" or "." not in struct, "Struct should contain ':' or be equal to '-': %s" % struct
-
-        if ":" in struct:
-            elem, attr = struct.split(":")
-        else:
-            elem = struct
-            attr = util.UNDEF
-        if struct and not struct == "-":
-            if elem not in structs:
-                structs[elem] = []
-                order.append(elem)
-            structs[elem].append((attr, n))
-    return [(elem, structs[elem]) for elem in order]
+    """Parse a list of annotation names (annotation:attribute) into a list of tuples."""
+    structs = OrderedDict()
+    for struct in structural_atts:
+        elem, _, attr = struct.partition(":")
+        if elem not in structs:
+            structs[elem] = []
+        if attr:
+            structs[elem].append(attr)
+    return [(elem, structs[elem]) for elem in structs]
 
 
 def cwb_escape(inname):
     """Replace dots with "-" for CWB compatibility."""
+    # From the CWB documentation: "By convention, all attribute names must be lowercase
+    # (more precisely, they may only contain the characters a-z, 0-9, -, and _, and may not start with a digit)"
     return re.sub(r"\.", "-", inname)
 
 
@@ -402,4 +410,4 @@ def truncateset(string, maxlength=4095, delimiter="|", affix="|", encoding="UTF-
         for i, value in enumerate(values):
             length += len(value.encode(encoding)) + 1
             if length > maxlength:
-                return util.cwbset(values[:i], delimiter, affix)
+                return util.misc.cwbset(values[:i], delimiter, affix)
