@@ -1,82 +1,125 @@
 """Create annotations from SALDO."""
 
 import itertools
-import logging
 import re
 from typing import List, Optional
 
-import sparv.util as util
-from sparv import Annotation, Config, Model, Output, annotator
-from sparv.modules.saldo.saldo_model import SaldoLexicon
+from sparv.api import Annotation, Config, Model, Output, annotator, get_logger, util
 
-log = logging.getLogger(__name__)
+from .saldo_model import SaldoLexicon
+
+logger = get_logger(__name__)
 
 # The minimum precision difference for two annotations to be considered equal
 PRECISION_DIFF = 0.01
 
 
+def preloader(models):
+    """Preload SALDO models."""
+    if not isinstance(models, list):
+        models = [models]
+    return {m.path.stem: SaldoLexicon(m.path) for m in models}
+
+
 @annotator("SALDO annotations", language=["swe"], config=[
     Config("saldo.model", default="saldo/saldo.pickle", description="Path to SALDO model"),
+    Config("saldo.delimiter", default=util.constants.DELIM, description="Character to put between ambiguous results"),
+    Config("saldo.affix", default=util.constants.AFFIX, description="Character to put before and after sets of results"),
     Config("saldo.precision", "",
-           description="Format string for appending precision to each value")
-])
+           description="Format string for appending precision to each value (e.g. ':%.3f')"),
+    Config("saldo.precision_filter", default="max",
+           description="Precision filter with possible values 'max' (only use the most probable annotations), "
+                       "'first' (only use the single most probable annotation), 'none' (use all annotations)"),
+    Config("saldo.min_precision", default=0.66,
+           description="Only use annotations with a probability score greater than or equal to this. "
+                       "0.25: part-of-speech does not match, 0.5: part-of-speech is missing, 0.66: part-of-speech "
+                       "matches, 0.75: morphosyntactic descriptor matches"),
+    Config("saldo.skip_multiword", default=False, description="Whether to disable annotation of multiword expressions"),
+    Config("saldo.max_mwe_gaps", default=1, description="Max amount of gaps allowed within a multiword expression"),
+    Config("saldo.allow_multiword_overlap", default=False,
+           description="Whether all multiword expressions may overlap with each other. "
+                       "If set to False, some cleanup is done."),
+    Config("saldo.word_separator", default="",
+           description="Character used to split the values of 'word' into several word variations"),
+], preloader=preloader, preloader_params=["models"], preloader_target="models_preloaded")
 def annotate(token: Annotation = Annotation("<token>"),
              word: Annotation = Annotation("<token:word>"),
              sentence: Annotation = Annotation("<sentence>"),
-             reference: Annotation = Annotation("<token>:misc.number_rel_<sentence>"),
-             out_sense: Output = Output("<token>:saldo.sense", cls="token:sense", description="SALDO identifier"),
-             out_lemgram: Output = Output("<token>:saldo.lemgram", description="SALDO lemgram"),
+             reference: Annotation = Annotation("<token:ref>"),
+             out_sense: Output = Output("<token>:saldo.sense", cls="token:sense", description="SALDO identifiers"),
+             out_lemgram: Output = Output("<token>:saldo.lemgram", cls="token:lemgram", description="SALDO lemgrams"),
              out_baseform: Output = Output("<token>:saldo.baseform", cls="token:baseform",
-                                           description="Baseform from SALDO"),
+                                           description="Baseforms from SALDO"),
              models: List[Model] = [Model("[saldo.model]")],
              msd: Optional[Annotation] = Annotation("<token:msd>"),
-             delimiter: str = util.DELIM,
-             affix: str = util.AFFIX,
+             delimiter: str = Config("saldo.delimiter"),
+             affix: str = Config("saldo.affix"),
              precision: str = Config("saldo.precision"),
-             precision_filter: str = "max",
-             min_precision: float = 0.66,
-             skip_multiword: bool = False,
-             allow_multiword_overlap: bool = False,
-             word_separator: str = "",
-             lexicons=None):
-    """Use the Saldo lexicon model (and optionally other older lexicons) to annotate pos-tagged words.
+             precision_filter: str = Config("saldo.precision_filter"),
+             min_precision: float = Config("saldo.min_precision"),
+             skip_multiword: bool = Config("saldo.skip_multiword"),
+             max_gaps: int = Config("saldo.max_mwe_gaps"),
+             allow_multiword_overlap: bool = Config("saldo.allow_multiword_overlap"),
+             word_separator: str = Config("saldo.word_separator"),
+             models_preloaded: Optional[dict] = None):
+    """Use the Saldo lexicon model to annotate msd-tagged words.
 
-    - token, word, msd, sentence, reference: existing annotations
-    - out_baseform, out_lemgram, out_sense: resulting annotations to be written
-    - models: a list of pickled lexica, typically the Saldo model (saldo.pickle)
-      and optional lexicons for older Swedish.
-    - delimiter: delimiter character to put between ambiguous results
-    - affix: an optional character to put before and after results
-    - precision: a format string for how to print the precision for each annotation, e.g. ":%.3f"
-      (use empty string for no precision)
-    - precision_filter: an optional filter, currently there are the following values:
-        max: only use the annotations that are most probable
-        first: only use the most probable annotation (or one of the most probable if more than one)
-        none: use all annotations
-    - min_precision: only use annotations with a probability score higher than this
-    - skip_multiword: set to True to disable multi word annotations
-    - allow_multiword_overlap: by default we do some cleanup among overlapping multi word annotations.
-      By setting this to True, all overlaps will be allowed.
-    - word_separator: an optional character used to split the values of "word" into several word variations
-    - lexicons: this argument cannot be set from the command line, but is used in the catapult.
-      This argument must be last.
+    Args:
+        token (Annotation): Input annotation with token spans. Defaults to Annotation("<token>").
+        word (Annotation): Input annotation with token strings. Defaults to Annotation("<token:word>").
+        sentence (Annotation): Input annotation with sentence spans. Defaults to Annotation("<sentence>").
+        reference (Annotation): Input annotation with token indices for each sentence.
+            Defaults to Annotation("<token:ref>").
+        out_sense (Output): Output annotation with senses from SALDO. Defaults to Output("<token>:saldo.sense").
+        out_lemgram (Output): Output annotation with lemgrams from SALDO. Defaults to Output("<token>:saldo.lemgram").
+        out_baseform (Output): Output annotation with baseforms from SALDO.
+            Defaults to Output("<token>:saldo.baseform").
+        models (List[Model]): A list of pickled lexicons, typically the SALDO model (saldo.pickle)
+            and optional lexicons for older Swedish. Defaults to [Model("[saldo.model]")].
+        msd (Annotation, optional): Input annotation with POS and morphological descriptions.
+            Defaults to Annotation("<token:msd>").
+        delimiter (str): Character to put between ambiguous results. Defaults to Config("saldo.delimiter").
+        affix (str): Character to put before and after sets of results. Defaults to Config("saldo.affix").
+        precision (str): Format string for appending precision to each value (e.g. ':%.3f'). use empty string for no
+            precision. Defaults to Config("saldo.precision").
+        precision_filter (str): Precision filter with values 'max' (only use the annotations that are most probable),
+            'first' (only use the most probable annotation(s)), 'none' (use all annotations)".
+            Defaults to Config("saldo.precision_filter").
+        min_precision (float): Only use annotations with a probability score higher than this.
+            Defaults to Config("saldo.min_precision").
+        skip_multiword (bool): Whether to disable annotation of multiword expressions.
+            Defaults to Config("saldo.skip_multiword").
+        max_gaps (int): Max amount of gaps allowed within a multiword expression. Defaults to Config("saldo.max_gaps").
+        allow_multiword_overlap (bool): Whether all multiword expressions may overlap with each other. If set to False,
+            some cleanup is done. Defaults to Config("saldo.allow_multiword_overlap").
+        word_separator (str): Character used to split the values of 'word' into several word variations.
+            Defaults to Config("saldo.word_separator").
+        models_preloaded (dict, optional): Preloaded models. Defaults to None.
     """
+    main(token=token, word=word, sentence=sentence, reference=reference, out_sense=out_sense, out_lemgram=out_lemgram,
+         out_baseform=out_baseform, models=models, msd=msd, delimiter=delimiter, affix=affix, precision=precision,
+         precision_filter=precision_filter, min_precision=min_precision, skip_multiword=skip_multiword,
+         max_gaps=max_gaps, allow_multiword_overlap=allow_multiword_overlap, word_separator=word_separator,
+         models_preloaded=models_preloaded)
+
+
+def main(token, word, sentence, reference, out_sense, out_lemgram, out_baseform, models, msd, delimiter, affix,
+         precision, precision_filter, min_precision, skip_multiword, max_gaps, allow_multiword_overlap, word_separator,
+         models_preloaded):
+    """Do SALDO annotations with models."""
     # Allow use of multiple lexicons
+    logger.progress()
     models_list = [(m.path.stem, m) for m in models]
-    if not lexicons:
+    if not models_preloaded:
         lexicon_list = [(name, SaldoLexicon(lex.path)) for name, lex in models_list]
-    # Use pre-loaded lexicons (from catapult)
+    # Use pre-loaded lexicons
     else:
         lexicon_list = []
         for name, _lex in models_list:
-            assert lexicons.get(name, None) is not None, "Lexicon %s not found!" % name
-            lexicon_list.append((name, lexicons[name]))
+            assert models_preloaded.get(name, None) is not None, "Lexicon %s not found!" % name
+            lexicon_list.append((name, models_preloaded[name]))
 
-    # Maximum number of gaps in multi-word units.
-    # TODO: Set to 0 for hist-mode? since many (most?) multi-word in the old lexicons are inseparable (half öre etc)
-    max_gaps = 1
-
-    # Combine annotation names i SALDO lexicon with out annotations
+    # Combine annotation names in SALDO lexicon with out annotations
     annotations = []
     if out_baseform:
         annotations.append((out_baseform, "gf"))
@@ -86,7 +129,7 @@ def annotate(token: Annotation = Annotation("<token>"),
         annotations.append((out_sense, "saldo"))
 
     if skip_multiword:
-        log.info("Skipping multi word annotations")
+        logger.info("Skipping multi word annotations")
 
     min_precision = float(min_precision)
 
@@ -98,11 +141,17 @@ def annotate(token: Annotation = Annotation("<token>"),
     ref_annotation = list(reference.read())
     if msd:
         msd_annotation = list(msd.read())
+    else:
+        msd_annotation = word.create_empty_attribute()
 
     sentences, orphans = sentence.get_children(token)
     sentences.append(orphans)
 
+    if orphans:
+        logger.warning(f"Found {len(orphans)} tokens not belonging to any sentence. These will not be annotated.")
+
     out_annotation = word.create_empty_attribute()
+    logger.progress(total=len(sentences) + 1)
 
     for sent in sentences:
         incomplete_multis = []  # [{annotation, words, [ref], is_particle, lastwordWasGap, numberofgaps}]
@@ -124,22 +173,23 @@ def annotate(token: Annotation = Annotation("<token>"),
                 thewords = [theword]
 
             # First use MSD tags to find the most probable single word annotations
-            ann_tags_words = find_single_word(thewords, lexicon_list, msdtag, precision, min_precision,
-                                              precision_filter, annotation_info)
+            ann_tags_words = _find_single_word(thewords, lexicon_list, msdtag, precision, min_precision,
+                                               precision_filter, annotation_info)
 
             # Find multi-word expressions
             if not skip_multiword:
-                find_multiword_expressions(incomplete_multis, complete_multis, thewords, ref, msdtag, max_gaps,
-                                           ann_tags_words, msd_annotation, sent, skip_pos_check)
+                _find_multiword_expressions(incomplete_multis, complete_multis, thewords, ref, msdtag, max_gaps,
+                                            ann_tags_words, msd_annotation, sent, skip_pos_check)
 
             # Loop to next token
+        logger.progress()
 
         if not allow_multiword_overlap:
             # Check that we don't have any unwanted overlaps
-            remove_unwanted_overlaps(complete_multis)
+            _remove_unwanted_overlaps(complete_multis)
 
         # Then save the rest of the multi word expressions in sentence_tokens
-        save_multiwords(complete_multis, sentence_tokens)
+        _save_multiwords(complete_multis, sentence_tokens)
 
         for tok in list(sentence_tokens.values()):
             out_annotation[tok["token_index"]] = _join_annotation(tok["annotations"], delimiter, affix)
@@ -147,15 +197,15 @@ def annotate(token: Annotation = Annotation("<token>"),
         # Loop to next sentence
 
     for out_annotation_obj, annotation_name in annotations:
-        out_annotation_obj.write([v.get(annotation_name, delimiter) for v in out_annotation])
+        out_annotation_obj.write([v.get(annotation_name, delimiter) if v is not None else None for v in out_annotation])
+    logger.progress()
 
 
 ################################################################################
 # Auxiliaries
 ################################################################################
 
-
-def find_single_word(thewords, lexicon_list, msdtag, precision, min_precision, precision_filter, annotation_info):
+def _find_single_word(thewords, lexicon_list, msdtag, precision, min_precision, precision_filter, annotation_info):
     ann_tags_words = []
 
     for w in thewords:
@@ -168,7 +218,7 @@ def find_single_word(thewords, lexicon_list, msdtag, precision, min_precision, p
             for a in lexicon.lookup(w):
                 annotation.append(a + (prefix,))
             ann_tags_words += annotation
-            # Set break if each word only gets annotations from first lexicon that has entry for word
+            # # Set break if each word only gets annotations from first lexicon that has entry for word
             # break
 
     annotation_precisions = [(get_precision(msdtag, msdtags), annotation, prefix)
@@ -176,7 +226,7 @@ def find_single_word(thewords, lexicon_list, msdtag, precision, min_precision, p
 
     if min_precision > 0:
         annotation_precisions = [x for x in annotation_precisions if x[0] >= min_precision]
-    annotation_precisions = normalize_precision(annotation_precisions)
+    annotation_precisions = _normalize_precision(annotation_precisions)
     annotation_precisions.sort(reverse=True, key=lambda x: x[0])
 
     if precision_filter and annotation_precisions:
@@ -214,8 +264,8 @@ def find_single_word(thewords, lexicon_list, msdtag, precision, min_precision, p
     return ann_tags_words
 
 
-def find_multiword_expressions(incomplete_multis, complete_multis, thewords, ref, msdtag, max_gaps, ann_tags_words,
-                               msd_annotation, sent, skip_pos_check):
+def _find_multiword_expressions(incomplete_multis, complete_multis, thewords, ref, msdtag, max_gaps, ann_tags_words,
+                                msd_annotation, sent, skip_pos_check):
     todelfromincomplete = []  # list to keep track of which expressions that have been completed
 
     for i, x in enumerate(incomplete_multis):
@@ -291,12 +341,20 @@ def find_multiword_expressions(incomplete_multis, complete_multis, thewords, ref
         incomplete_multis.extend(looking_for)
 
 
-def remove_unwanted_overlaps(complete_multis):
+def _remove_unwanted_overlaps(complete_multis):
+    """Remove certain overlapping MWEs if they have identical POS (remove 'a' if 'b1 a1 b2 a2' or 'a1 b1 ab2')."""
     remove = set()
     for ai, a in enumerate(complete_multis):
+        # For historical texts: Since we allow many words for one token (spelling variations) we must make sure that
+        # two words of an MWE are not made up by two variants of one token. That is, that the same ref ID is not
+        # used twice in an MWE.
+        if len(set(a[0])) != len(a[0]):
+            remove.add(ai)
+            continue
         for b in complete_multis:
             # Check if both are of same POS
-            if not a == b and re.search(r"\.(\w\w?)m?\.", a[1]["lem"][0]).groups()[0] == re.search(r"\.(\w\w?)m?\.", b[1]["lem"][0]).groups()[0]:
+            if not a == b and re.search(r"\.(\w\w?)m?\.", a[1]["lem"][0]).groups()[0] == re.search(
+                    r"\.(\w\w?)m?\.", b[1]["lem"][0]).groups()[0]:
                 if b[0][0] < a[0][0] < b[0][-1] < a[0][-1]:
                     # A case of b1 a1 b2 a2. Remove a.
                     remove.add(ai)
@@ -308,7 +366,7 @@ def remove_unwanted_overlaps(complete_multis):
         del complete_multis[a]
 
 
-def save_multiwords(complete_multis, sentence_tokens):
+def _save_multiwords(complete_multis, sentence_tokens):
     for c in complete_multis:
         first = True
         first_ref = ""
@@ -323,8 +381,9 @@ def save_multiwords(complete_multis, sentence_tokens):
 
 
 def _join_annotation(annotation, delimiter, affix):
-    seen = set()
-    return dict([(a, affix + delimiter.join(b for b in annotation[a] if b not in seen and not seen.add(b)) + affix) for a in annotation])
+    """Convert annotations into cwb sets with unique values."""
+    return dict([(a, util.misc.cwbset(list(dict.fromkeys(annotation[a])), delimiter=delimiter, affix=affix))
+                 for a in annotation])
 
 
 def get_precision(msd, msdtags):
@@ -341,7 +400,7 @@ def get_precision(msd, msdtags):
             0.25)
 
 
-def normalize_precision(annotations):
+def _normalize_precision(annotations):
     """Normalize the rankings in the annotation list so that the sum is 1."""
     total_precision = sum(prec for (prec, _annotation, prefix) in annotations)
     return [(prec / total_precision, annotation, prefix) for (prec, annotation, prefix) in annotations]
