@@ -1,14 +1,13 @@
 """Segmentation mostly based on NLTK."""
 
-import logging
+import inspect
 import pickle
 import re
 from typing import Optional
 
 import nltk
 
-import sparv.util as util
-from sparv import Annotation, Config, Model, ModelOutput, Output, Text, annotator, modelbuilder
+from sparv.api import Annotation, Config, Model, ModelOutput, Output, Text, annotator, get_logger, modelbuilder, util
 from sparv.modules.saldo.saldo_model import split_triple
 
 try:
@@ -16,7 +15,7 @@ try:
 except ImportError:
     pass
 
-log = logging.getLogger(__name__)
+logger = get_logger(__name__)
 
 
 @annotator("Automatic tokenization", config=[
@@ -32,7 +31,7 @@ def tokenize(text: Text = Text(),
              out: Output = Output("segment.token", cls="token", description="Token segments"),
              chunk: Annotation = Annotation("[segment.token_chunk]"),
              segmenter: str = Config("segment.token_segmenter"),
-             existing_segments: Optional[str] = Config("segment.existing_tokens"),
+             existing_segments: Optional[Annotation] = Annotation("[segment.existing_tokens]"),
              model: Optional[Model] = Model("[segment.tokenizer_config]"),
              token_list: Optional[Model] = Model("[segment.token_list]")):
     """Tokenize text."""
@@ -51,7 +50,7 @@ def sentence(text: Text = Text(),
              out: Output = Output("segment.sentence", cls="sentence", description="Sentence segments"),
              chunk: Optional[Annotation] = Annotation("[segment.sentence_chunk]"),
              segmenter: str = Config("segment.sentence_segmenter"),
-             existing_segments: Optional[str] = Config("segment.existing_sentences"),
+             existing_segments: Optional[Annotation] = Annotation("[segment.existing_sentences]"),
              model: Optional[Model] = Model("[segment.sentence_model]")):
     """Split text into sentences."""
     do_segmentation(text=text, out=out, chunk=chunk, segmenter=segmenter, existing_segments=existing_segments,
@@ -68,31 +67,36 @@ def paragraph(text: Text = Text(),
               out: Output = Output("segment.paragraph", cls="paragraph", description="Paragraph segments"),
               chunk: Optional[Annotation] = Annotation("[segment.paragraph_chunk]"),
               segmenter: str = Config("segment.paragraph_segmenter"),
-              existing_segments: Optional[str] = Config("segment.existing_paragraphs"),
+              existing_segments: Optional[Annotation] = Annotation("[segment.existing_paragraphs]"),
               model: Optional[Model] = None):
     """Split text into paragraphs."""
     do_segmentation(text=text, out=out, chunk=chunk, segmenter=segmenter, existing_segments=existing_segments,
                     model=model)
 
 
-def do_segmentation(text: Text, out: Output, segmenter, chunk: Optional[Annotation] = None, existing_segments=None,
-                    model: Optional[Model] = None, token_list: Optional[Model] = None):
+def do_segmentation(text: Text, out: Output, segmenter, chunk: Optional[Annotation] = None,
+                    existing_segments: Optional[Annotation] = None, model: Optional[Model] = None,
+                    token_list: Optional[Model] = None):
     """Segment all chunks (e.g. sentences) into smaller "tokens" (e.g. words), and annotate them as "element" (e.g. w).
 
     Segmentation is done by the given "segmenter"; some segmenters take
     an extra argument which is a pickled "model" object.
     """
-    segmenter_args = []
-    if model:
-        if model.path.suffix in ["pickle", "pkl"]:
-            with open(model, "rb") as M:
+    assert segmenter in SEGMENTERS, "Available segmenters: %s" % ", ".join(sorted(SEGMENTERS))
+    segmenter = SEGMENTERS[segmenter]
+
+    segmenter_args = {}
+    if model and "model" in inspect.getfullargspec(segmenter).args:
+        if model.path.suffix in [".pickle", ".pkl"]:
+            with open(model.path, "rb") as M:
                 model_arg = pickle.load(M, encoding="UTF-8")
         else:
             model_arg = str(model.path)
-        segmenter_args.append(model_arg)
-    assert segmenter in SEGMENTERS, "Available segmenters: %s" % ", ".join(sorted(SEGMENTERS))
-    segmenter = SEGMENTERS[segmenter]
-    segmenter = segmenter(*segmenter_args)
+        segmenter_args["model"] = model_arg
+    if token_list and "token_list" in inspect.getfullargspec(segmenter).args:
+        segmenter_args["token_list"] = str(token_list.path)
+
+    segmenter = segmenter(**segmenter_args)
     assert hasattr(segmenter, "span_tokenize"), "Segmenter needs a 'span_tokenize' method: %r" % segmenter
 
     corpus_text = text.read()
@@ -121,10 +125,12 @@ def do_segmentation(text: Text, out: Output, segmenter, chunk: Optional[Annotati
                 chunk_start = segment_end
                 chunk_spans[n] = (chunk_start, chunk_end)
         chunk_spans.sort()
-        log.info("Reorganized into %d chunks" % len(chunk_spans))
+        logger.info("Reorganized into %d chunks" % len(chunk_spans))
     else:
         segments = []
 
+    logger.progress(total=len(chunk_spans) + 1)
+
     # Now we can segment each chunk span into tokens
     for start, end in chunk_spans:
         for spanstart, spanend in segmenter.span_tokenize(corpus_text[start:end]):
@@ -133,9 +139,11 @@ def do_segmentation(text: Text, out: Output, segmenter, chunk: Optional[Annotati
             if corpus_text[spanstart:spanend].strip():
                 span = (spanstart, spanend)
                 segments.append(span)
+        logger.progress()
 
     segments.sort()
     out.write(segments)
+    logger.progress()
 
 
 @modelbuilder("Model for PunktSentenceTokenizer", language=["swe"])
@@ -161,8 +169,8 @@ def build_tokenlist(saldo_model: Model = Model("saldo/saldo.pickle"),
     """Build a list of words from a SALDO model, to help BetterWordTokenizer."""
     segmenter_args = []
     if model:
-        if model.path.suffix in ["pickle", "pkl"]:
-            with open(model, "rb") as m:
+        if model.path.suffix in [".pickle", ".pkl"]:
+            with open(model.path, "rb") as m:
                 model_arg = pickle.load(m)
         else:
             model_arg = model.path
@@ -191,23 +199,23 @@ def build_tokenlist(saldo_model: Model = Model("saldo/saldo.pickle"),
 
 ######################################################################
 
-def train_punkt_segmenter(textfiles, modelfile, encoding=util.UTF8, protocol=-1):
+def train_punkt_segmenter(textfiles, modelfile, encoding=util.constants.UTF8, protocol=-1):
     """Train a Punkt sentence tokenizer."""
     if isinstance(textfiles, str):
         textfiles = textfiles.split()
 
-    log.info("Reading files")
+    logger.info("Reading files")
     text = u""
     for filename in textfiles:
         with open(filename, encoding=encoding) as stream:
             text += stream.read()
-    log.info("Training model")
+    logger.info("Training model")
     trainer = nltk.tokenize.PunktTrainer(text, verbose=True)
-    log.info("Saving pickled model")
+    logger.info("Saving pickled model")
     params = trainer.get_params()
     with open(modelfile, "wb") as stream:
         pickle.dump(params, stream, protocol=protocol)
-    log.info("OK")
+    logger.info("OK")
 
 
 ######################################################################
@@ -234,15 +242,12 @@ class PunctuationTokenizer(nltk.RegexpTokenizer):
         """Tokenize s and return list with tokens."""
         result = []
         spans = nltk.RegexpTokenizer.span_tokenize(self, s)
-        first = True
         temp = [0, 0]
 
         for start, _ in spans:
-            if not first:
-                temp[1] = start
-                result.append(tuple(temp))
+            temp[1] = start
+            result.append(tuple(temp))
             temp[0] = start
-            first = False
 
         temp[1] = len(s)
         result.append(tuple(temp))
@@ -305,7 +310,7 @@ class BetterWordTokenizer:
                         try:
                             key, val = line.strip().split(None, 1)
                         except ValueError as e:
-                            log.error("Error parsing configuration file: %s", line)
+                            logger.error("Error parsing configuration file: %s", line)
                             raise e
                         key = key[:-1]
 
@@ -335,7 +340,8 @@ class BetterWordTokenizer:
                 self._word_tokenize_fmt %
                 {
                     "tokens": ("(?:" + "|".join(self.patterns["tokens"]) + ")|") if self.patterns["tokens"] else "",
-                    "abbrevs": ("(?:" + "|".join(re.escape(a + ".") for a in self.abbreviations) + ")|") if self.abbreviations else "",
+                    "abbrevs": ("(?:" + "|".join(
+                        re.escape(a + ".") for a in self.abbreviations) + ")|") if self.abbreviations else "",
                     "misc": "|".join(self.patterns["misc"]),
                     "number": self.patterns["number"],
                     "within": self.patterns["within"],
@@ -377,6 +383,14 @@ class BetterWordTokenizer:
             begin += len(w)
 
 
+class PunktSentenceTokenizer(nltk.PunktSentenceTokenizer):
+    """A simple subclass of nltk.PunktSentenceTokenizer to add the required 'model' parameter."""
+
+    def __init__(self, model):
+        """Call __init__ of superclass."""
+        super().__init__(model)
+
+
 class CRFTokenizer:
     """Tokenization based on Conditional Random Fields.
 
@@ -426,7 +440,7 @@ class FSVParagraphSplitter:
 SEGMENTERS = dict(whitespace=nltk.WhitespaceTokenizer,
                   linebreaks=LinebreakTokenizer,
                   blanklines=nltk.BlanklineTokenizer,
-                  punkt_sentence=nltk.PunktSentenceTokenizer,
+                  punkt_sentence=PunktSentenceTokenizer,
                   punctuation=PunctuationTokenizer,
                   better_word=BetterWordTokenizer,
                   crf_tokenizer=CRFTokenizer,
